/*
 *  Microfeed - Backend for accessing feed-based services
 *  Copyright (C) 2009 Henrik Hedberg <henrik.hedberg@innologies.fi>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <microfeed-provider/microfeeddatabase.h>
#include <microfeed-common/microfeedmisc.h>
#include <microfeed-common/microfeedthread.h>
#include <microfeed-common/microfeedstore.h>
#include <microfeed-common/microfeedobject.h>
#include <microfeed-common/microfeedprotocol.h>

#include <errno.h>
#include <string.h>
#include <db.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

struct _MicrofeedDatabaseEnvironment {
	MicrofeedObject object;
	char* public_name;
	char* private_name;
	char* directory;
	DB_ENV* db_env;
	MicrofeedStore* databases;
};

struct _MicrofeedDatabase {
	MicrofeedObject object;
	MicrofeedDatabaseEnvironment* database_environment;
	char* name;
	MicrofeedDatabaseCompareFunction compare_function;
	DB* db;
	MicrofeedDatabaseIndex* indices;
	MicrofeedDatabaseIterator* iterators;
};

struct _MicrofeedDatabaseIndex {
	MicrofeedDatabaseIndex* next;
	
	MicrofeedDatabase* database;
	char* name;
	DB* db;
	MicrofeedDatabaseIndexFunction index_function;
	MicrofeedDatabaseCompareFunction compare_function;
};

struct _MicrofeedDatabaseIterator {
	MicrofeedDatabaseIterator* previous;
	MicrofeedDatabaseIterator* next;

	MicrofeedDatabase* database;
	DBC* db_cursor;
	DBT key;
	DBT data;
	void (*next_function)(MicrofeedDatabaseIterator* iterator);
	DB* index_db;
	DBT index_key;
	int backwards : 1;
};

static void database_environment_free_function(MicrofeedDatabaseEnvironment* database_environment);
static void database_free_function(MicrofeedDatabase* database);

static void lock_environment_for_writing(MicrofeedDatabaseEnvironment* database_environment);
static void unlock_environment(MicrofeedDatabaseEnvironment* database_environment);
static int key_compare_function(DB *db, const DBT *dbt1, const DBT *dbt2);
static int dup_compare_function(DB *db, const DBT *dbt1, const DBT *dbt2);
static int index_callback(DB* secondary, const DBT* key, const DBT* data, DBT* result);
static void remove_data(MicrofeedDatabase* database, DB* db, const void* key, const size_t key_size);
static void remove_data_range(MicrofeedDatabase* database, DB* db, const void* start_key, const size_t start_key_size, const void* end_key, const size_t end_key_size);
static void iterator_free(MicrofeedDatabaseIterator* iterator);
static void database_iterator_next(MicrofeedDatabaseIterator* iterator);
static void database_index_iterator_next(MicrofeedDatabaseIterator* iterator);
	
static MicrofeedClass microfeed_database_environment_class = {
	"MicrofeedDatabaseEnvironment",
	(MicrofeedObjectFreeFunction)database_environment_free_function
};

static MicrofeedClass microfeed_database_class = {
	"MicrofeedDatabase",
	(MicrofeedObjectFreeFunction)database_free_function
};
	
static MicrofeedDatabaseEnvironment* new_database_environment(const char* public_name, const char* private_name, const char* directory) {
	MicrofeedDatabaseEnvironment* database_environment = NULL;
	DB_ENV* db_env;
	int error;

	if (!(error = db_env_create(&db_env, 0)) &&
	    !(error = db_env->open(db_env, directory, DB_CREATE | DB_THREAD | DB_PRIVATE | DB_INIT_CDB | DB_INIT_MPOOL, 0))) {
		database_environment = microfeed_object_new(&microfeed_database_environment_class, MicrofeedDatabaseEnvironment);
		database_environment->public_name = strdup(public_name);
		database_environment->private_name = strdup(private_name);
		database_environment->directory = strdup(directory);
		database_environment->db_env = db_env;
		database_environment->databases = microfeed_store_new_sorted_weak_references((MicrofeedStoreCompareKeysFunction)strcmp,
			                                                	             (MicrofeedStoreGetKeyFunction)microfeed_database_get_name);
	} else {
		fprintf(stderr, "ERROR: Failed to create database environment: %s\n", db_strerror(error));
	}
	
	return database_environment;

}

MicrofeedDatabaseEnvironment* microfeed_database_environment_new(const char* name, const char* directory) {
	MicrofeedDatabaseEnvironment* database_environment = NULL;
	char buffer[1024];
	char* source_path;
	char* destination_path;
	int fd;
	struct flock lock;

	snprintf(buffer, 1024, "%s%c%ld", name, MICROFEED_PUBLISHER_IDENTIFIER_SEPARATOR_CHAR, (long)getpid());
	if (microfeed_util_create_directory_recursively(directory)) {
		source_path = microfeed_util_string_concatenate(directory, "/", name, NULL);
		destination_path = microfeed_util_string_concatenate(directory, "/", buffer, NULL);
		if ((fd = open(source_path, O_RDWR)) >= 0) {
			lock.l_type = F_RDLCK;
			lock.l_whence = SEEK_SET;
			lock.l_start = 0;
			lock.l_len = 0;
			if (fcntl(fd, F_GETLK, &lock) == 0) {
				/* Note: file_copy opens and closes the source file, and thus, releases the lock. */
				if (microfeed_util_file_copy(source_path, destination_path)) {
					if (!(database_environment = new_database_environment(name, buffer, directory))) {
						unlink(destination_path);
					}
				}
			}
			/* This, at least, releases the lock. */
			close(fd);
		} else {
			database_environment = new_database_environment(name, buffer, directory);
		}
		microfeed_memory_free(source_path);
		microfeed_memory_free(destination_path);
	}

	return database_environment;
}

static void database_environment_free_function(MicrofeedDatabaseEnvironment* database_environment) {
	char* source_path;
	char* destination_path;
	int error;
	
	microfeed_store_clean_weak_references(database_environment->databases);
	microfeed_assert(microfeed_store_get_size(database_environment->databases) == 0);

	if (!(error = database_environment->db_env->close(database_environment->db_env, 0))) {
		source_path = microfeed_util_string_concatenate(database_environment->directory, "/", database_environment->private_name, NULL);
		destination_path = microfeed_util_string_concatenate(database_environment->directory, "/", database_environment->public_name, NULL);
		unlink(destination_path);
		link(source_path, destination_path);
		unlink(source_path);
		microfeed_memory_free(source_path);
		microfeed_memory_free(destination_path);
	} else {
		fprintf(stderr, "ERROR: Could not close database environment: %s\n", db_strerror(error));
	}

	free(database_environment->public_name);
	free(database_environment->private_name);
	free(database_environment->directory);
	microfeed_store_free(database_environment->databases);
	database_environment->db_env = NULL;
}

MicrofeedDatabase* microfeed_database_environment_get_database(MicrofeedDatabaseEnvironment* database_environment, const char* name, MicrofeedDatabaseCompareFunction compare_function, int create) {
	MicrofeedDatabase* database = NULL;
	DB* db;
	int error;
	
	lock_environment_for_writing(database_environment);

	if (!(database = microfeed_store_get(database_environment->databases, name, MicrofeedDatabase)) &&
	    !(error = db_create(&db, database_environment->db_env, 0))) {
	    	if ((!compare_function || !(error = db->set_bt_compare(db, key_compare_function))) &&
		    !(error = db->open(db, NULL, database_environment->private_name, name, DB_BTREE, DB_THREAD | (create ? DB_CREATE : 0), 0))) {
			database = microfeed_object_new(&microfeed_database_class, MicrofeedDatabase);
			database->database_environment = microfeed_object_ref(database_environment, MicrofeedDatabaseEnvironment);
			database->name = strdup(name);
			database->compare_function = compare_function;
			database->db = db;
			database->indices = NULL;
			database->iterators = NULL;
			database->db->app_private = database;

			microfeed_store_insert(database_environment->databases, database);
		} else {
			if ((error = db->close(db, 0))) {
				fprintf(stderr, "ERROR: Could not closed database that is not opened: %s\n", db_strerror(error));
			}
		}
	}

	unlock_environment(database_environment);

	return database;
}

MicrofeedDatabaseIndex* microfeed_database_get_index(MicrofeedDatabase* database, const char* name, MicrofeedDatabaseIndexFunction index_function, MicrofeedDatabaseCompareFunction compare_function) {
	MicrofeedDatabaseIndex* database_index;
	char* full_name;
	DB* db;
	int error;
	
	lock_environment_for_writing(database->database_environment);

	for (database_index = database->indices; database_index; database_index = database_index->next) {
		if (!strcmp(database_index->name, name)) {
			break;
		}
	}
	if (!database_index) {
		full_name = microfeed_util_string_concatenate(database->name, ":microfeed.index:", name, NULL);
		if (!(error = db_create(&db, database->database_environment->db_env, 0)) &&
		    (!compare_function || !(error = db->set_dup_compare(db, dup_compare_function))) &&
		    !(error = db->open(db, NULL, database->database_environment->private_name, full_name, DB_BTREE, DB_CREATE | DB_THREAD, 0))) {
			if (!(error = db->associate(database->db, NULL, db, index_callback, DB_CREATE))) { // | DB_IMMUTABLE_KEY)) {
				database_index = microfeed_memory_allocate(MicrofeedDatabaseIndex);
				database_index->db = db;
				database_index->database = database;
				database_index->name = strdup(name);
				database_index->index_function = index_function;
				database_index->compare_function = compare_function;		
				database_index->db->app_private = database_index;

				database_index->next = database->indices;
				database->indices = database_index;
			} else {
				fprintf(stderr, "ERROR: Failed to associate secondary index: %s\n", db_strerror(error));
				db->close(db, 0);
			}
		} else {
			fprintf(stderr, "ERROR: Failed to create index database: %s\n", db_strerror(error));
		}
		free(full_name);
	}

	unlock_environment(database->database_environment);
	
	return database_index;	
}

static void database_free_function(MicrofeedDatabase* database) {
	MicrofeedDatabaseIndex* database_index;
	MicrofeedDatabaseIndex* next_database_index;
	int error;

	microfeed_assert(!database->iterators);

	for (database_index = database->indices; database_index; database_index = next_database_index) {
		next_database_index = database_index->next;
		if ((error = database_index->db->close(database_index->db, 0))) {
			fprintf(stderr, "ERROR: Failed to close database index: %s\n", database_index->name);
		}
		free(database_index->name);
		microfeed_memory_free(database_index);
	}	
	if ((error = database->db->close(database->db, 0))) {
		fprintf(stderr, "ERROR: Failed to close database: %s\n", database->name);
	}
	database->db = NULL;

	microfeed_object_unref(database->database_environment, MicrofeedDatabaseEnvironment);
	database->database_environment = NULL;
	free(database->name);
	database->name = NULL;
}

const char* microfeed_database_get_name(MicrofeedDatabase* database) {
	
	return database->name;
}

int microfeed_database_get_data(MicrofeedDatabase* database, const void* key, size_t key_size, void** data, size_t* data_size) {
	int retvalue = 0;
	DBT dbt_key;
	DBT dbt_data;
	
	microfeed_object_lock(database, MicrofeedDatabase);

	memset(&dbt_key, 0, sizeof(DBT));
	memset(&dbt_data, 0, sizeof(DBT));
	dbt_key.data = (void*)key;
	dbt_key.size = key_size;
	dbt_data.flags = DB_DBT_MALLOC;
	
	if (!database->db->get(database->db, NULL, &dbt_key, &dbt_data, 0)) {
		*data = dbt_data.data;
		*data_size = dbt_data.size;
		retvalue = 1;
	}
	
	microfeed_object_unlock(database, MicrofeedDatabase);
	
	return retvalue;
}

int microfeed_database_get_data_partial(MicrofeedDatabase* database, const void* key, size_t key_size, void* data, size_t* data_size, size_t offset) {
	int retvalue = 0;
	DBT dbt_key;
	DBT dbt_data;
	
	microfeed_object_lock(database, MicrofeedDatabase);

	memset(&dbt_key, 0, sizeof(DBT));
	memset(&dbt_data, 0, sizeof(DBT));
	dbt_key.data = (void*)key;
	dbt_key.size = key_size;
	dbt_data.data = (void*)data;
	dbt_data.ulen = *data_size;
	dbt_data.flags = DB_DBT_USERMEM | DB_DBT_PARTIAL;
	dbt_data.dlen = *data_size;
	dbt_data.doff = offset;
	
	if (!database->db->get(database->db, NULL, &dbt_key, &dbt_data, 0)) {
		*data_size = dbt_data.size;
		retvalue = 1;
	}
	
	microfeed_object_unlock(database, MicrofeedDatabase);
	
	return retvalue;
}


MicrofeedDatabaseIterator* microfeed_database_iterate(MicrofeedDatabase* database, const void* start_key, const size_t start_key_size, int backwards) {
	MicrofeedDatabaseIterator* iterator;

	microfeed_object_lock(database, MicrofeedDatabase);

	iterator = microfeed_memory_allocate(MicrofeedDatabaseIterator);
	iterator->database = database;
	iterator->next_function = database_iterator_next;
	iterator->backwards = backwards;
	iterator->index_db = NULL;
	memset(&iterator->key, 0, sizeof(DBT));
	memset(&iterator->data, 0, sizeof(DBT));
	iterator->key.flags = DB_DBT_REALLOC;
	iterator->data.flags = DB_DBT_REALLOC;
	iterator->data.data = malloc(1);
	iterator->data.size = 1;

	if (start_key) {
		iterator->key.data = malloc(start_key_size);
		memcpy(iterator->key.data, start_key, start_key_size);
		iterator->key.size = start_key_size;
	}
	iterator->next_function(iterator);

	if (database->iterators) {
		database->iterators->previous = iterator;
		iterator->next = database->iterators;
	} else {
		iterator->next = NULL;
	}
	iterator->previous = NULL;
	database->iterators = iterator;

	microfeed_object_unlock(database, MicrofeedDatabase);

	return iterator;
}

/*
void print(const char* text, const char* data, const size_t size) {
	size_t i;
	
	printf("%s: ", text);
	for (i = 0; i < size; i++) {
		if ((data[i] >= 'a' && data[i] <= 'z') || (data[i] >= 'A' && data[i] <= 'Z') || (data[i] >= '0' && data[i] <= '9')) {
			printf("%c", data[i]);
		} else {
			printf("_");
		}
	}
	printf("\n");
}
*/

void microfeed_database_replace_data(MicrofeedDatabase* database, const void* key, const size_t key_size, const void* data, const size_t data_size) {
	DBT dbt_key;
	DBT dbt_data;
	int error;
	
	lock_environment_for_writing(database->database_environment);

	memset(&dbt_key, 0, sizeof(DBT));
	memset(&dbt_data, 0, sizeof(DBT));
	dbt_key.data = (void*)key;
	dbt_key.size = key_size;
	dbt_data.data = (void*)data;
	dbt_data.size = data_size;
	if ((error = database->db->put(database->db, NULL, &dbt_key, &dbt_data, 0))) {
		fprintf(stderr, "DB->put failed: %s\n", db_strerror(error));
	}

	unlock_environment(database->database_environment);
}

void microfeed_database_replace_data_partial(MicrofeedDatabase* database, const void* key, size_t key_size, const void* data, size_t data_size, size_t offset) {
	DBT dbt_key;
	DBT dbt_data;
	int error;
	
	lock_environment_for_writing(database->database_environment);

	memset(&dbt_key, 0, sizeof(DBT));
	memset(&dbt_data, 0, sizeof(DBT));
	dbt_key.data = (void*)key;
	dbt_key.size = key_size;
	dbt_data.data = (void*)data;
	dbt_data.size = data_size;
	dbt_data.flags = DB_DBT_PARTIAL;
	dbt_data.dlen = data_size;
	dbt_data.doff = offset;
	if ((error = database->db->put(database->db, NULL, &dbt_key, &dbt_data, 0))) {
		fprintf(stderr, "DB->put failed in replace_data_partial: %s\n", db_strerror(error));
	}

	unlock_environment(database->database_environment);
}

void microfeed_database_remove_data(MicrofeedDatabase* database, const void* key, const size_t key_size) {
	remove_data(database, database->db, key, key_size);
}

void microfeed_database_remove_data_range(MicrofeedDatabase* database, const void* start_key, const size_t start_key_size, const void* end_key, const size_t end_key_size) {
	remove_data_range(database, database->db, start_key, start_key_size, end_key, end_key_size);
}

MicrofeedDatabase* microfeed_database_index_get_database(MicrofeedDatabaseIndex* database_index) {
	
	return database_index->database;
}

const char* microfeed_database_index_get_name(MicrofeedDatabaseIndex* database_index) {
	
	return database_index->name;
}

int microfeed_database_index_get_data(MicrofeedDatabaseIndex* database_index, const void* index_key, const size_t index_key_size, void** key, size_t* key_size, void** data, size_t* data_size) {
	int retvalue = 0;
	DBT dbt_index_key;
	DBT dbt_key;
	DBT dbt_data;

	microfeed_object_lock(database_index->database, MicrofeedDatabase);

	memset(&dbt_index_key, 0, sizeof(DBT));
	memset(&dbt_key, 0, sizeof(DBT));
	memset(&dbt_data, 0, sizeof(DBT));
	dbt_index_key.data = (void*)index_key;
	dbt_index_key.size = index_key_size;
	dbt_key.flags = DB_DBT_MALLOC;
	dbt_data.flags = DB_DBT_MALLOC;
	
	if (!database_index->db->pget(database_index->db, NULL, &dbt_index_key, &dbt_key, &dbt_data, 0)) {
		*key = (char*)dbt_key.data;
		*key_size = dbt_key.size;
		*data = dbt_data.data;
		*data_size = dbt_data.size;
		retvalue = 1;
	}
	
	microfeed_object_unlock(database_index->database, MicrofeedDatabase);
	
	return retvalue;
	
}

MicrofeedDatabaseIterator* microfeed_database_index_iterate(MicrofeedDatabaseIndex* database_index, const void* start_key, const size_t start_key_size, int backwards) {
	MicrofeedDatabaseIterator* iterator;

	microfeed_object_lock(database_index->database, MicrofeedDatabase);

	iterator = microfeed_memory_allocate(MicrofeedDatabaseIterator);
	iterator->database = database_index->database;
	iterator->next_function = database_index_iterator_next;
	iterator->backwards = backwards;
	iterator->index_db = database_index->db;
	memset(&iterator->key, 0, sizeof(DBT));
	memset(&iterator->data, 0, sizeof(DBT));
	memset(&iterator->index_key, 0, sizeof(DBT));
	iterator->key.flags = DB_DBT_REALLOC;
	iterator->data.flags = DB_DBT_REALLOC;
	iterator->index_key.flags = DB_DBT_REALLOC;
	iterator->data.data = malloc(1);
	iterator->data.size = 1;

	if (start_key) {
		iterator->index_key.data = malloc(start_key_size);
		memcpy(iterator->index_key.data, start_key, start_key_size);
		iterator->index_key.size = start_key_size;
	}
	iterator->next_function(iterator);

	if (database_index->database->iterators) {
		database_index->database->iterators->previous = iterator;
		iterator->next = database_index->database->iterators;
	} else {
		iterator->next = NULL;
	}
	iterator->previous = NULL;
	database_index->database->iterators = iterator;

	microfeed_object_unlock(database_index->database, MicrofeedDatabase);

	return iterator;
}
	
void microfeed_database_index_remove_data(MicrofeedDatabaseIndex* database_index, const void* key, const size_t key_size) {
	remove_data(database_index->database, database_index->db, key, key_size);
}

void microfeed_database_index_remove_data_range(MicrofeedDatabaseIndex* database_index, const void* start_key, const size_t start_key_size, const void* end_key, const size_t end_key_size) {
	remove_data_range(database_index->database, database_index->db, start_key, start_key_size, end_key, end_key_size);	
}

void microfeed_database_iterator_free(MicrofeedDatabaseIterator* iterator) {
	microfeed_object_lock(iterator->database, MicrofeedDatabase);

	if (iterator->next) {
		iterator->next->previous = iterator->previous;
	}
	if (iterator->previous) {
		iterator->previous->next = iterator->next;
	} else if (iterator->database) {
		iterator->database->iterators = iterator->next;
	}

	microfeed_object_unlock(iterator->database, MicrofeedDatabase);

	iterator_free(iterator);
	iterator->database = NULL;
	microfeed_memory_free(iterator);
}
	
int microfeed_database_iterator_get(MicrofeedDatabaseIterator* iterator, const void** key, size_t* key_size, const void** data, size_t* data_size) {
	int success = 0;

	microfeed_object_lock(iterator->database, MicrofeedDatabase);
	
	if (iterator->data.data) {
		*key = iterator->key.data;
		*key_size = iterator->key.size;
		*data = iterator->data.data;
		*data_size = iterator->data.size;
		success = 1;
	}
	
	microfeed_object_unlock(iterator->database, MicrofeedDatabase);

	return success;
}

void microfeed_database_iterator_next(MicrofeedDatabaseIterator* iterator) {
	microfeed_object_lock(iterator->database, MicrofeedDatabase);

	iterator->next_function(iterator);
	
	microfeed_object_unlock(iterator->database, MicrofeedDatabase);
}

static void remove_data(MicrofeedDatabase* database, DB* db, const void* key, const size_t key_size) {
	DBT dbt_key;
	int error;
	
	lock_environment_for_writing(database->database_environment);

	memset(&dbt_key, 0, sizeof(DBT));
	dbt_key.data = (void*)key;
	dbt_key.size = key_size;

	if ((error = db->del(db, NULL, &dbt_key, 0))) {
		fprintf(stderr, "ERROR: Failed to delete data from database: %s\n", db_strerror(error));
	}

	unlock_environment(database->database_environment);
}

static void remove_data_range(MicrofeedDatabase* database, DB* db, const void* start_key, const size_t start_key_size, const void* end_key, const size_t end_key_size) {
	DBC* db_cursor;
	DBT dbt_key;
	DBT dbt_data;

	lock_environment_for_writing(database->database_environment);	

	if (!db->cursor(db, NULL, &db_cursor, DB_WRITECURSOR)) {
		memset(&dbt_key, 0, sizeof(DBT));
		memset(&dbt_data, 0, sizeof(DBT));
		dbt_key.flags = DB_DBT_REALLOC;
		dbt_data.flags = DB_DBT_REALLOC;
		if (start_key) {
			dbt_key.data = malloc(start_key_size);
			memcpy(dbt_key.data, start_key, start_key_size);
			dbt_key.size = start_key_size;
		}
		if ((start_key && !db_cursor->c_get(db_cursor, &dbt_key, &dbt_data, DB_SET_RANGE)) || (!start_key && !db_cursor->c_get(db_cursor, &dbt_key, &dbt_data, DB_FIRST))) {
			do {
				db_cursor->c_del(db_cursor, 0);
				if (end_key && end_key_size == dbt_key.size && memcmp(end_key, dbt_key.data, end_key_size) <= 0) {
					break;
				}
			} while (!db_cursor->c_get(db_cursor, &dbt_key, &dbt_data, DB_NEXT));
		}
		db_cursor->c_close(db_cursor);
		free(dbt_key.data);
		free(dbt_data.data);
	}

	unlock_environment(database->database_environment);
	
}

static void iterator_free(MicrofeedDatabaseIterator* iterator) {
	if (iterator->db_cursor) {
		iterator->db_cursor->c_close(iterator->db_cursor);
		iterator->db_cursor = NULL;
	}
	free(iterator->key.data);
	free(iterator->data.data);
	free(iterator->index_key.data);
	iterator->key.data = NULL;
	iterator->data.data = NULL;
	iterator->index_key.data = NULL;
}

static void database_iterator_next(MicrofeedDatabaseIterator* iterator) {
	char* previous_key = NULL;
	
	if (iterator->data.data) {
		if (!iterator->db_cursor) {
			if (iterator->key.data) {
				previous_key = malloc(iterator->key.size);
				memcpy(previous_key, iterator->key.data, iterator->key.size);		
				if (iterator->database->db->cursor(iterator->database->db, NULL, &iterator->db_cursor, 0) ||
				    iterator->db_cursor->c_get(iterator->db_cursor, &iterator->key, &iterator->data, DB_SET_RANGE) ||
				    (!strcmp(iterator->key.data, previous_key) && iterator->db_cursor->c_get(iterator->db_cursor, &iterator->key, &iterator->data, (iterator->backwards ? DB_PREV : DB_NEXT)))) {
					iterator_free(iterator);
				}
				free(previous_key);
			} else {
				if (iterator->database->db->cursor(iterator->database->db, NULL, &iterator->db_cursor, 0) ||
				    iterator->db_cursor->c_get(iterator->db_cursor, &iterator->key, &iterator->data, (iterator->key.data ? DB_SET_RANGE : (iterator->backwards ? DB_LAST : DB_FIRST)))) {
					iterator_free(iterator);
				}
			}
		} else {
			if (iterator->db_cursor->c_get(iterator->db_cursor, &iterator->key, &iterator->data, (iterator->backwards ? DB_PREV : DB_NEXT))) {
				iterator_free(iterator);
			}
		}
	}
}

static void database_index_iterator_next(MicrofeedDatabaseIterator* iterator) {
	char* previous_key = NULL;
	
	if (iterator->data.data) {
		if (!iterator->db_cursor) {
			if (iterator->index_key.data) {
				previous_key = malloc(iterator->index_key.size);
				memcpy(previous_key, iterator->index_key.data, iterator->index_key.size);
				if (iterator->index_db->cursor(iterator->index_db, NULL, &iterator->db_cursor, 0) ||
				    iterator->db_cursor->c_pget(iterator->db_cursor, &iterator->index_key, &iterator->key, &iterator->data, DB_SET_RANGE) ||
				    (!strcmp(iterator->index_key.data, previous_key) && iterator->db_cursor->c_pget(iterator->db_cursor, &iterator->index_key, &iterator->key, &iterator->data, (iterator->backwards ? DB_PREV : DB_NEXT)))) {
					iterator_free(iterator);
				}
				free(previous_key);
			} else {
				if (iterator->index_db->cursor(iterator->index_db, NULL, &iterator->db_cursor, 0) ||
				    iterator->db_cursor->c_pget(iterator->db_cursor, &iterator->index_key, &iterator->key, &iterator->data, (iterator->index_key.data ? DB_SET_RANGE : (iterator->backwards ? DB_LAST : DB_FIRST)))) {
					iterator_free(iterator);
				}
			}
		} else {
			if (iterator->db_cursor->c_pget(iterator->db_cursor, &iterator->index_key, &iterator->key, &iterator->data, (iterator->backwards ? DB_PREV : DB_NEXT))) {
				iterator_free(iterator);
			}
		}
	}
}

static void lock_environment_for_writing(MicrofeedDatabaseEnvironment* database_environment) {
	MicrofeedStoreIterator* store_iterator;
	MicrofeedDatabase* database;
	MicrofeedDatabaseIterator* database_iterator;

	microfeed_object_lock(database_environment, MicrofeedDatabaseEnvironment);

	/* TODO: Must be cleaned before since microfeed_store_iterator_get has a bug! */
	microfeed_store_clean_weak_references(database_environment->databases);
	for (store_iterator = microfeed_store_iterate(database_environment->databases, NULL);
	     (database = microfeed_store_iterator_get(store_iterator, MicrofeedDatabase));
	     microfeed_store_iterator_next(store_iterator)) {
		microfeed_object_lock(database, MicrofeedDatabase);
		for (database_iterator = database->iterators; database_iterator; database_iterator = database_iterator->next) {
			if (database_iterator->data.data && database_iterator->db_cursor) {
				database_iterator->db_cursor->c_close(database_iterator->db_cursor);
				database_iterator->db_cursor = NULL;
			}
		}
	}
	microfeed_store_iterator_free(store_iterator);
}

static void unlock_environment(MicrofeedDatabaseEnvironment* database_environment) {
	MicrofeedStoreIterator* store_iterator;
	MicrofeedDatabase* database;

	/* TODO: Must be cleaned before since microfeed_store_iterator_get has a bug! */
	microfeed_store_clean_weak_references(database_environment->databases);
	for (store_iterator = microfeed_store_iterate(database_environment->databases, NULL);
	     (database = microfeed_store_iterator_get(store_iterator, MicrofeedDatabase));
	     microfeed_store_iterator_next(store_iterator)) {
		microfeed_object_unlock(database, MicrofeedDatabase);
	}
	microfeed_store_iterator_free(store_iterator);

	microfeed_object_unlock(database_environment, MicrofeedDatabaseEnvironment);
}

static int index_callback(DB* db, const DBT* key, const DBT* data, DBT* result) {
	int retvalue = EINVAL;
	MicrofeedDatabaseIndex* database_index;
	void* index_key;
	size_t index_key_size;
	
	if ((database_index = (MicrofeedDatabaseIndex*)db->app_private)) {
		database_index->index_function(key->data, key->size, data->data, data->size, &index_key, &index_key_size);
		result->data = index_key;
		result->size = index_key_size;
		retvalue = 0;
	}
	
	return retvalue;
}

static int key_compare_function(DB *db, const DBT *dbt1, const DBT *dbt2) {
	int retvalue = 0;
	MicrofeedDatabase* database;
	
	if ((database = (MicrofeedDatabase*)db->app_private)) {
		retvalue = database->compare_function(dbt1->data, dbt1->size, dbt2->data, dbt2->size);	
	}

	return retvalue;
}

static int dup_compare_function(DB *db, const DBT *dbt1, const DBT *dbt2) {
	int retvalue = 0;
	MicrofeedDatabaseIndex* database_index;
	
	if ((database_index = (MicrofeedDatabaseIndex*)db->app_private)) {
		retvalue = database_index->compare_function(dbt1->data, dbt1->size, dbt2->data, dbt2->size);	
	}

	return retvalue;
}
